@@ -38,8 +38,19 @@
 
 #include "intel_batchbuffer.h"
 
-#define DRIVER_ANY 0x1
-#define DRIVER_INTEL (0x1 << 1)
+#define DRIVER_INTEL	(1 << 0)
+#define DRIVER_VC4	(1 << 1)
+#define DRIVER_VGEM	(1 << 2)
+#define DRIVER_VIRTIO	(1 << 3)
+#define DRIVER_AMDGPU	(1 << 4)
+/*
+ * Exclude DRVER_VGEM from DRIVER_ANY since if you run on a system
+ * with vgem as well as a supported driver, you can end up with a
+ * near-100% skip rate if you don't explicitly specify the device,
+ * depending on device-load ordering.
+ */
+#define DRIVER_ANY 	~(DRIVER_VGEM)
+
 
 #ifdef ANDROID
 #if (!(defined HAVE_MMAP64)) && (!(defined __x86_64__))
@@ -81,6 +92,10 @@ int __drm_open_driver(int chipset);
 
 void gem_quiescent_gpu(int fd);
 
+void igt_require_intel(int fd);
+
+bool is_i915_device(int fd);
+
 /**
  * do_or_die:
  * @x: command
@@ -101,7 +116,7 @@ void gem_quiescent_gpu(int fd);
  * successfully executed.
  */
 #define do_ioctl(fd, ioc, ioc_data) do { \
-	igt_assert_eq(drmIoctl((fd), (ioc), (ioc_data)), 0); \
+	igt_assert_eq(igt_ioctl((fd), (ioc), (ioc_data)), 0); \
 	errno = 0; \
 } while (0)
 
@@ -116,7 +131,7 @@ void gem_quiescent_gpu(int fd);
  * returning a particular value in errno.
  */
 #define do_ioctl_err(fd, ioc, ioc_data, err) do { \
-	igt_assert_eq(drmIoctl((fd), (ioc), (ioc_data)), -1); \
+	igt_assert_eq(igt_ioctl((fd), (ioc), (ioc_data)), -1); \
 	igt_assert_eq(errno, err); \
 	errno = 0; \
 } while (0)
