@@ -24,9 +24,7 @@
  *    Robert Foss <robert.foss@collabora.com>
  */
 
-#ifndef ANDROID
-#define _GNU_SOURCE
-#else
+#ifdef HAVE_LIBGEN_H
 #include <libgen.h>
 #endif
 #include <fcntl.h>
@@ -41,6 +39,13 @@
 #include "drmtest.h"
 #include "ioctl_wrappers.h"
 
+/**
+ * SECTION:sw_sync
+ * @short_description: Software sync (fencing) support library
+ * @title: SW Sync
+ * @include: sw_sync.h
+ */
+
 struct int_sync_create_fence_data {
 	__u32	value;
 	char	name[32];
