@@ -110,7 +110,10 @@ static void from_mmap(int fd, uint64_t size, enum mode mode)
 	uint64_t max, i;
 	int retry = 2;
 
-	intel_require_memory(1, size, CHECK_RAM);
+	/* Worst case is that the kernel has to copy the entire incoming
+	 * reloc[], so double the memory requirements.
+	 */
+	intel_require_memory(2, size, CHECK_RAM);
 
 	memset(&obj, 0, sizeof(obj));
 	obj.handle = gem_create(fd, 4096);
@@ -142,7 +145,7 @@ static void from_mmap(int fd, uint64_t size, enum mode mode)
 		reloc_handle = gem_create(fd, size);
 		relocs = gem_mmap__wc(fd, reloc_handle, 0, size, PROT_WRITE);
 		gem_set_domain(fd, reloc_handle,
-			       I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_GTT);
+			       I915_GEM_DOMAIN_WC, I915_GEM_DOMAIN_WC);
 		gem_close(fd, reloc_handle);
 		break;
 	}
@@ -187,6 +190,8 @@ static void from_gpu(int fd)
 	uint32_t reloc_handle;
 	uint64_t value;
 
+	igt_require(gem_can_store_dword(fd, 0));
+
 	memset(&obj, 0, sizeof(obj));
 	obj.handle = gem_create(fd, 4096);
 	gem_write(fd, obj.handle, 0, &bbe, sizeof(bbe));
@@ -228,11 +233,6 @@ static void from_gpu(int fd)
 	munmap(relocs, 4096);
 }
 
-static bool ignore_engine(int gen, unsigned engine)
-{
-	return gen == 6 && (engine & ~(3<<13)) == I915_EXEC_BSD;
-}
-
 static void check_bo(int fd, uint32_t handle)
 {
 	uint32_t *map;
@@ -259,12 +259,12 @@ static void active(int fd, unsigned engine)
 	nengine = 0;
 	if (engine == -1) {
 		for_each_engine(fd, engine) {
-			if (!ignore_engine(gen, engine))
+			if (gem_can_store_dword(fd, engine))
 				engines[nengine++] = engine;
 		}
 	} else {
 		igt_require(gem_has_ring(fd, engine));
-		igt_require(!ignore_engine(gen, engine));
+		igt_require(gem_can_store_dword(fd, engine));
 		engines[nengine++] = engine;
 	}
 	igt_require(nengine);
@@ -373,6 +373,8 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 			if (before == I915_GEM_DOMAIN_CPU)
 				wc = gem_mmap__cpu(fd, obj.handle, 0, OBJSZ, PROT_WRITE);
 			else if (before == I915_GEM_DOMAIN_GTT)
+				wc = gem_mmap__gtt(fd, obj.handle, OBJSZ, PROT_WRITE);
+			else if (before == I915_GEM_DOMAIN_WC)
 				wc = gem_mmap__wc(fd, obj.handle, 0, OBJSZ, PROT_WRITE);
 			else
 				igt_assert(0);
@@ -386,7 +388,7 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 		}
 
 		if (flags & ACTIVE) {
-			spin = igt_spin_batch_new(fd, I915_EXEC_DEFAULT, obj.handle);
+			spin = igt_spin_batch_new(fd, 0, I915_EXEC_DEFAULT, obj.handle);
 			if (!(flags & HANG))
 				igt_spin_batch_set_timeout(spin, NSEC_PER_SEC/100);
 			igt_assert(gem_bo_busy(fd, obj.handle));
@@ -400,6 +402,8 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 			if (after == I915_GEM_DOMAIN_CPU)
 				wc = gem_mmap__cpu(fd, obj.handle, 0, OBJSZ, PROT_READ);
 			else if (after == I915_GEM_DOMAIN_GTT)
+				wc = gem_mmap__gtt(fd, obj.handle, OBJSZ, PROT_READ);
+			else if (after == I915_GEM_DOMAIN_WC)
 				wc = gem_mmap__wc(fd, obj.handle, 0, OBJSZ, PROT_READ);
 			else
 				igt_assert(0);
@@ -427,7 +431,6 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 		} else {
 			trash = obj.handle;
 			obj.handle = gem_create(fd, OBJSZ);
-			obj.offset = -1;
 			gem_write(fd, obj.handle, 0, &bbe, sizeof(bbe));
 			reloc.target_handle = obj.handle;
 		}
@@ -438,20 +441,20 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 			if (before == I915_GEM_DOMAIN_CPU)
 				wc = gem_mmap__cpu(fd, obj.handle, 0, OBJSZ, PROT_WRITE);
 			else if (before == I915_GEM_DOMAIN_GTT)
+				wc = gem_mmap__gtt(fd, obj.handle, OBJSZ, PROT_WRITE);
+			else if (before == I915_GEM_DOMAIN_WC)
 				wc = gem_mmap__wc(fd, obj.handle, 0, OBJSZ, PROT_WRITE);
 			else
 				igt_assert(0);
 			gem_set_domain(fd, obj.handle, before, before);
-			offset = -1;
-			memcpy(wc + reloc_offset, &offset, sizeof(offset));
+			memcpy(wc + reloc_offset, &reloc.presumed_offset, sizeof(reloc.presumed_offset));
 			munmap(wc, OBJSZ);
 		} else {
-			offset = -1;
-			gem_write(fd, obj.handle, reloc_offset, &offset, sizeof(offset));
+			gem_write(fd, obj.handle, reloc_offset, &reloc.presumed_offset, sizeof(reloc.presumed_offset));
 		}
 
 		if (flags & ACTIVE) {
-			spin = igt_spin_batch_new(fd, I915_EXEC_DEFAULT, obj.handle);
+			spin = igt_spin_batch_new(fd, 0, I915_EXEC_DEFAULT, obj.handle);
 			if (!(flags & HANG))
 				igt_spin_batch_set_timeout(spin, NSEC_PER_SEC/100);
 			igt_assert(gem_bo_busy(fd, obj.handle));
@@ -465,6 +468,8 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 			if (after == I915_GEM_DOMAIN_CPU)
 				wc = gem_mmap__cpu(fd, obj.handle, 0, OBJSZ, PROT_READ);
 			else if (after == I915_GEM_DOMAIN_GTT)
+				wc = gem_mmap__gtt(fd, obj.handle, OBJSZ, PROT_READ);
+			else if (after == I915_GEM_DOMAIN_WC)
 				wc = gem_mmap__wc(fd, obj.handle, 0, OBJSZ, PROT_READ);
 			else
 				igt_assert(0);
@@ -491,6 +496,124 @@ static void basic_reloc(int fd, unsigned before, unsigned after, unsigned flags)
 	gem_close(fd, obj.handle);
 }
 
+static inline uint64_t sign_extend(uint64_t x, int index)
+{
+	int shift = 63 - index;
+	return (int64_t)(x << shift) >> shift;
+}
+
+static uint64_t gen8_canonical_address(uint64_t address)
+{
+	return sign_extend(address, 47);
+}
+
+static void basic_range(int fd, unsigned flags)
+{
+	struct drm_i915_gem_relocation_entry reloc[128];
+	struct drm_i915_gem_exec_object2 obj[128];
+	struct drm_i915_gem_execbuffer2 execbuf;
+	uint64_t address_mask = has_64b_reloc(fd) ? ~(uint64_t)0 : ~(uint32_t)0;
+	uint64_t gtt_size = gem_aperture_size(fd);
+	const uint32_t bbe = MI_BATCH_BUFFER_END;
+	igt_spin_t *spin = NULL;
+	int count, n;
+
+	igt_require(gem_has_softpin(fd));
+
+	for (count = 12; gtt_size >> (count + 1); count++)
+		;
+
+	count -= 12;
+
+	memset(obj, 0, sizeof(obj));
+	memset(reloc, 0, sizeof(reloc));
+	memset(&execbuf, 0, sizeof(execbuf));
+
+	n = 0;
+	for (int i = 0; i <= count; i++) {
+		obj[n].handle = gem_create(fd, 4096);
+		obj[n].offset = (1ull << (i + 12)) - 4096;
+		obj[n].offset = gen8_canonical_address(obj[n].offset);
+		obj[n].flags = EXEC_OBJECT_PINNED | EXEC_OBJECT_SUPPORTS_48B_ADDRESS;
+		gem_write(fd, obj[n].handle, 0, &bbe, sizeof(bbe));
+		execbuf.buffers_ptr = to_user_pointer(&obj[n]);
+		execbuf.buffer_count = 1;
+		if (__gem_execbuf(fd, &execbuf))
+			continue;
+
+		igt_debug("obj[%d] handle=%d, address=%llx\n",
+			  n, obj[n].handle, (long long)obj[n].offset);
+
+		reloc[n].offset = 8 * (n + 1);
+		reloc[n].target_handle = obj[n].handle;
+		reloc[n].read_domains = I915_GEM_DOMAIN_INSTRUCTION;
+		reloc[n].presumed_offset = -1;
+		n++;
+	}
+	for (int i = 1; i < count; i++) {
+		obj[n].handle = gem_create(fd, 4096);
+		obj[n].offset = 1ull << (i + 12);
+		obj[n].offset = gen8_canonical_address(obj[n].offset);
+		obj[n].flags = EXEC_OBJECT_PINNED | EXEC_OBJECT_SUPPORTS_48B_ADDRESS;
+		gem_write(fd, obj[n].handle, 0, &bbe, sizeof(bbe));
+		execbuf.buffers_ptr = to_user_pointer(&obj[n]);
+		execbuf.buffer_count = 1;
+		if (__gem_execbuf(fd, &execbuf))
+			continue;
+
+		igt_debug("obj[%d] handle=%d, address=%llx\n",
+			  n, obj[n].handle, (long long)obj[n].offset);
+
+		reloc[n].offset = 8 * (n + 1);
+		reloc[n].target_handle = obj[n].handle;
+		reloc[n].read_domains = I915_GEM_DOMAIN_INSTRUCTION;
+		reloc[n].presumed_offset = -1;
+		n++;
+	}
+	igt_require(n);
+
+	obj[n].handle = gem_create(fd, 4096);
+	obj[n].relocs_ptr = to_user_pointer(reloc);
+	obj[n].relocation_count = n;
+	gem_write(fd, obj[n].handle, 0, &bbe, sizeof(bbe));
+
+	execbuf.buffers_ptr = to_user_pointer(obj);
+	execbuf.buffer_count = n + 1;
+
+	if (flags & ACTIVE) {
+		spin = igt_spin_batch_new(fd, 0, 0, obj[n].handle);
+		if (!(flags & HANG))
+			igt_spin_batch_set_timeout(spin, NSEC_PER_SEC/100);
+		igt_assert(gem_bo_busy(fd, obj[n].handle));
+	}
+
+	gem_execbuf(fd, &execbuf);
+	igt_spin_batch_free(fd, spin);
+
+	for (int i = 0; i < n; i++) {
+		uint64_t offset;
+
+		offset = ~reloc[i].presumed_offset & address_mask;
+		gem_read(fd, obj[n].handle, reloc[i].offset,
+			 &offset, has_64b_reloc(fd) ? 8 : 4);
+
+		igt_debug("obj[%d] handle=%d, offset=%llx, found=%llx, presumed=%llx\n",
+			  i, obj[i].handle,
+			  (long long)obj[i].offset,
+			  (long long)offset,
+			  (long long)reloc[i].presumed_offset);
+
+		igt_assert_eq_u64(obj[i].offset, offset);
+		if (reloc[i].presumed_offset == -1)
+			igt_warn("reloc.presumed_offset == -1\n");
+		else
+			igt_assert_eq_u64(reloc[i].presumed_offset, offset);
+	}
+
+	for (int i = 0; i <= n; i++)
+		gem_close(fd, obj[i].handle);
+}
+
 static void basic_softpin(int fd)
 {
 	struct drm_i915_gem_exec_object2 obj[2];
@@ -533,12 +656,19 @@ igt_main
 	} modes[] = {
 		{ "cpu", I915_GEM_DOMAIN_CPU, I915_GEM_DOMAIN_CPU },
 		{ "gtt", I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_GTT },
+		{ "wc", I915_GEM_DOMAIN_WC, I915_GEM_DOMAIN_WC },
 		{ "cpu-gtt", I915_GEM_DOMAIN_CPU, I915_GEM_DOMAIN_GTT },
 		{ "gtt-cpu", I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_CPU },
+		{ "cpu-wc", I915_GEM_DOMAIN_CPU, I915_GEM_DOMAIN_WC },
+		{ "wc-cpu", I915_GEM_DOMAIN_WC, I915_GEM_DOMAIN_CPU },
+		{ "gtt-wc", I915_GEM_DOMAIN_GTT, I915_GEM_DOMAIN_WC },
+		{ "wc-gtt", I915_GEM_DOMAIN_WC, I915_GEM_DOMAIN_GTT },
 		{ "cpu-read", I915_GEM_DOMAIN_CPU, 0 },
 		{ "gtt-read", I915_GEM_DOMAIN_GTT, 0 },
+		{ "wc-read", I915_GEM_DOMAIN_WC, 0 },
 		{ "write-cpu", 0, I915_GEM_DOMAIN_CPU },
 		{ "write-gtt", 0, I915_GEM_DOMAIN_GTT },
+		{ "write-wc", 0, I915_GEM_DOMAIN_WC },
 		{ "write-read", 0, 0 },
 		{ },
 	}, *m;
@@ -574,8 +704,17 @@ igt_main
 				igt_subtest_f("%s%s%s",
 					      f->basic ? "basic-" : "",
 					      m->name,
-					      f->name)
+					      f->name) {
+					if ((m->before | m->after) & I915_GEM_DOMAIN_WC)
+						igt_require(gem_mmap__has_wc(fd));
 					basic_reloc(fd, m->before, m->after, f->flags);
+				}
+			}
+
+			if (!(f->flags & NORELOC)) {
+				igt_subtest_f("%srange%s",
+					      f->basic ? "basic-" : "", f->name)
+					basic_range(fd, f->flags);
 			}
 
 			igt_fixture {
@@ -595,8 +734,10 @@ igt_main
 			from_mmap(fd, size, MEM | RO);
 		igt_subtest_f("cpu-%u", find_last_set(size) - 1)
 			from_mmap(fd, size, CPU);
-		igt_subtest_f("wc-%u", find_last_set(size) - 1)
+		igt_subtest_f("wc-%u", find_last_set(size) - 1) {
+			igt_require(gem_mmap__has_wc(fd));
 			from_mmap(fd, size, WC);
+		}
 		igt_subtest_f("gtt-%u", find_last_set(size) - 1)
 			from_mmap(fd, size, GTT);
 	}
