@@ -30,6 +30,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
+#include <poll.h>
 
 #ifndef DRM_CAP_CURSOR_WIDTH
 #define DRM_CAP_CURSOR_WIDTH 0x8
@@ -45,6 +46,11 @@ struct plane_parms {
 
 #define hweight32 __builtin_popcount
 
+/* globals for fence support */
+int *timeline;
+pthread_t *thread;
+int *seqno;
+
 static void
 run_primary_test(igt_display_t *display, enum pipe pipe, igt_output_t *output)
 {
@@ -88,9 +94,34 @@ run_primary_test(igt_display_t *display, enum pipe pipe, igt_output_t *output)
 	igt_remove_fb(display->drm_fd, &fb);
 }
 
+static void *fence_inc_thread(void *arg)
+{
+	int t = *((int *) arg);
+
+	pthread_detach(pthread_self());
+
+	usleep(5000);
+	sw_sync_timeline_inc(t, 1);
+	return NULL;
+}
+
+static void configure_fencing(igt_plane_t *plane)
+{
+	int i, fd, ret;
+
+	i = plane->index;
+
+	seqno[i]++;
+	fd = sw_sync_timeline_create_fence(timeline[i], seqno[i]);
+	igt_plane_set_fence_fd(plane, fd);
+	close(fd);
+	ret = pthread_create(&thread[i], NULL, fence_inc_thread, &timeline[i]);
+	igt_assert_eq(ret, 0);
+}
+
 static void
 wm_setup_plane(igt_display_t *display, enum pipe pipe,
-	       uint32_t mask, struct plane_parms *parms)
+	       uint32_t mask, struct plane_parms *parms, bool fencing)
 {
 	igt_plane_t *plane;
 
@@ -107,6 +138,9 @@ wm_setup_plane(igt_display_t *display, enum pipe pipe,
 			continue;
 		}
 
+		if (fencing)
+			configure_fencing(plane);
+
 		igt_plane_set_fb(plane, parms[i].fb);
 		igt_fb_set_size(parms[i].fb, plane, parms[i].width, parms[i].height);
 		igt_plane_set_size(plane, parms[i].width, parms[i].height);
@@ -135,7 +169,7 @@ static bool skip_on_unsupported_nonblocking_modeset(igt_display_t *display)
 
 	/* Force the next state to update all crtc's, to synchronize with the nonblocking modeset. */
 	for_each_pipe(display, pipe)
-		display->pipes[pipe].mode_changed = true;
+		igt_pipe_refresh(display, pipe, false);
 
 	return false;
 }
@@ -152,6 +186,7 @@ static drmEventContext drm_events = {
 
 enum transition_type {
 	TRANSITION_PLANES,
+	TRANSITION_AFTER_FREE,
 	TRANSITION_MODESET,
 	TRANSITION_MODESET_DISABLE,
 };
@@ -181,6 +216,7 @@ static void set_sprite_wh(igt_display_t *display, enum pipe pipe,
 
 static void setup_parms(igt_display_t *display, enum pipe pipe,
 			const drmModeModeInfo *mode,
+			struct igt_fb *primary_fb,
 			struct igt_fb *argb_fb,
 			struct igt_fb *sprite_fb,
 			struct plane_parms *parms)
@@ -203,7 +239,7 @@ static void setup_parms(igt_display_t *display, enum pipe pipe,
 		int i = plane->index;
 
 		if (plane->type == DRM_PLANE_TYPE_PRIMARY) {
-			parms[i].fb = plane->fb;
+			parms[i].fb = primary_fb;
 			parms[i].width = mode->hdisplay;
 			parms[i].height = mode->vdisplay;
 		} else if (plane->type == DRM_PLANE_TYPE_CURSOR) {
@@ -240,7 +276,7 @@ retry:
 		set_sprite_wh(display, pipe, parms, sprite_fb,
 			      alpha, sprite_width, sprite_height);
 
-		wm_setup_plane(display, pipe, (1 << n_planes) - 1, parms);
+		wm_setup_plane(display, pipe, (1 << n_planes) - 1, parms, false);
 		ret = igt_display_try_commit_atomic(display, DRM_MODE_ATOMIC_TEST_ONLY | DRM_MODE_ATOMIC_ALLOW_MODESET, NULL);
 
 		if (ret == -EINVAL) {
@@ -298,10 +334,6 @@ retry:
 		 sprite_width, sprite_height, alpha);
 }
 
-int *timeline;
-pthread_t *thread;
-int *seqno;
-
 static void prepare_fencing(igt_display_t *display, enum pipe pipe)
 {
 	igt_plane_t *plane;
@@ -333,57 +365,37 @@ static void unprepare_fencing(igt_display_t *display, enum pipe pipe)
 	free(seqno);
 }
 
-static void *fence_inc_thread(void *arg)
-{
-	int t = *((int *) arg);
-
-	pthread_detach(pthread_self());
-
-	usleep(5000);
-	sw_sync_timeline_inc(t, 1);
-	return NULL;
-}
-
-static void configure_fencing(igt_display_t *display, enum pipe pipe)
-{
-	igt_plane_t *plane;
-	int i, fd, ret;
-
-	for_each_plane_on_pipe(display, pipe, plane) {
-
-		if (!plane->fb)
-			continue;
-
-		i = plane->index;
-
-		seqno[i]++;
-		fd = sw_sync_timeline_create_fence(timeline[i], seqno[i]);
-		igt_plane_set_fence_fd(plane, fd);
-		close(fd);
-		ret = pthread_create(&thread[i], NULL, fence_inc_thread, &timeline[i]);
-		igt_assert_eq(ret, 0);
-	}
-}
-
-static void clear_fencing(igt_display_t *display, enum pipe pipe)
-{
-	igt_plane_t *plane;
-
-	for_each_plane_on_pipe(display, pipe, plane)
-		igt_plane_set_fence_fd(plane, -1);
-}
-
 static void atomic_commit(igt_display_t *display, enum pipe pipe, unsigned int flags, void *data, bool fencing)
 {
-	if (fencing) {
-		configure_fencing(display, pipe);
+	if (fencing)
 		igt_pipe_request_out_fence(&display->pipes[pipe]);
-	}
 
 	igt_display_commit_atomic(display, flags, data);
+}
 
-	if (fencing)
-		clear_fencing(display, pipe);
+static int fd_completed(int fd)
+{
+	struct pollfd pfd = { fd, POLLIN };
+	int ret;
+
+	ret = poll(&pfd, 1, 0);
+	igt_assert(ret >= 0);
+	return ret;
+}
+
+static void wait_for_transition(igt_display_t *display, enum pipe pipe, bool nonblocking, bool fencing)
+{
+	if (fencing) {
+		int fence_fd = display->pipes[pipe].out_fence_fd;
+
+		igt_assert_neq(fd_completed(fence_fd), nonblocking);
+
+		igt_assert(sync_fence_wait(fence_fd, 30000) == 0);
+	} else {
+		igt_assert_neq(fd_completed(display->drm_fd), nonblocking);
+
+		drmHandleEvent(display->drm_fd, &drm_events);
+	}
 }
 
 /*
@@ -402,14 +414,17 @@ run_transition_test(igt_display_t *display, enum pipe pipe, igt_output_t *output
 	struct igt_fb fb, argb_fb, sprite_fb;
 	drmModeModeInfo *mode, override_mode;
 	igt_plane_t *plane;
-	uint32_t iter_max = 1 << display->pipes[pipe].n_planes, i;
-	struct plane_parms parms[display->pipes[pipe].n_planes];
+	igt_pipe_t *pipe_obj = &display->pipes[pipe];
+	uint32_t iter_max = 1 << pipe_obj->n_planes, i;
+	struct plane_parms parms[pipe_obj->n_planes];
 	bool skip_test = false;
-	unsigned flags = DRM_MODE_PAGE_FLIP_EVENT;
+	unsigned flags = 0;
 	int ret;
 
 	if (fencing)
 		prepare_fencing(display, pipe);
+	else
+		flags |= DRM_MODE_PAGE_FLIP_EVENT;
 
 	if (nonblocking)
 		flags |= DRM_MODE_ATOMIC_NONBLOCK;
@@ -426,7 +441,7 @@ run_transition_test(igt_display_t *display, enum pipe pipe, igt_output_t *output
 
 	igt_output_set_pipe(output, pipe);
 
-	wm_setup_plane(display, pipe, 0, NULL);
+	wm_setup_plane(display, pipe, 0, NULL, false);
 
 	if (flags & DRM_MODE_ATOMIC_ALLOW_MODESET) {
 		skip_test = nonblocking && skip_on_unsupported_nonblocking_modeset(display);
@@ -442,7 +457,7 @@ run_transition_test(igt_display_t *display, enum pipe pipe, igt_output_t *output
 
 	igt_display_commit2(display, COMMIT_ATOMIC);
 
-	setup_parms(display, pipe, mode, &argb_fb, &sprite_fb, parms);
+	setup_parms(display, pipe, mode, &fb, &argb_fb, &sprite_fb, parms);
 
 	/*
 	 * In some configurations the tests may not run to completion with all
@@ -450,13 +465,13 @@ run_transition_test(igt_display_t *display, enum pipe pipe, igt_output_t *output
 	 * planes to fix this
 	 */
 	while (1) {
-		wm_setup_plane(display, pipe, iter_max - 1, parms);
+		wm_setup_plane(display, pipe, iter_max - 1, parms, false);
 
 		if (fencing)
-			igt_pipe_request_out_fence(&display->pipes[pipe]);
+			igt_pipe_request_out_fence(pipe_obj);
 
 		ret = igt_display_try_commit_atomic(display, DRM_MODE_ATOMIC_TEST_ONLY | DRM_MODE_ATOMIC_ALLOW_MODESET, NULL);
-		if (ret != -EINVAL || display->pipes[pipe].n_planes < 3)
+		if (ret != -EINVAL || pipe_obj->n_planes < 3)
 			break;
 
 		ret = 0;
@@ -480,40 +495,69 @@ run_transition_test(igt_display_t *display, enum pipe pipe, igt_output_t *output
 			igt_skip("Cannot run tests without proper size sprite planes\n");
 	}
 
+	igt_display_commit2(display, COMMIT_ATOMIC);
+
+	if (type == TRANSITION_AFTER_FREE) {
+		int fence_fd = -1;
+
+		wm_setup_plane(display, pipe, 0, parms, fencing);
+
+		atomic_commit(display, pipe, flags, (void *)(unsigned long)0, fencing);
+		if (fencing) {
+			fence_fd = pipe_obj->out_fence_fd;
+			pipe_obj->out_fence_fd = -1;
+		}
+
+		/* force planes to be part of commit */
+		for_each_plane_on_pipe(display, pipe, plane)
+			igt_plane_set_position(plane, 0, 0);
+
+		igt_display_commit2(display, COMMIT_ATOMIC);
+
+		if (fence_fd != -1) {
+			igt_assert(fd_completed(fence_fd));
+			close(fence_fd);
+		} else {
+			igt_assert(fd_completed(display->drm_fd));
+			wait_for_transition(display, pipe, false, fencing);
+		}
+		goto cleanup;
+	}
+
 	for (i = 0; i < iter_max; i++) {
 		igt_output_set_pipe(output, pipe);
 
-		wm_setup_plane(display, pipe, i, parms);
+		wm_setup_plane(display, pipe, i, parms, fencing);
 
 		atomic_commit(display, pipe, flags, (void *)(unsigned long)i, fencing);
-		drmHandleEvent(display->drm_fd, &drm_events);
+		wait_for_transition(display, pipe, nonblocking, fencing);
 
 		if (type == TRANSITION_MODESET_DISABLE) {
 			igt_output_set_pipe(output, PIPE_NONE);
 
-			wm_setup_plane(display, pipe, 0, parms);
+			wm_setup_plane(display, pipe, 0, parms, fencing);
 
 			atomic_commit(display, pipe, flags, (void *) 0UL, fencing);
-			drmHandleEvent(display->drm_fd, &drm_events);
+			wait_for_transition(display, pipe, nonblocking, fencing);
 		} else {
 			uint32_t j;
 
 			/* i -> i+1 will be done when i increases, can be skipped here */
 			for (j = iter_max - 1; j > i + 1; j--) {
-				wm_setup_plane(display, pipe, j, parms);
+				wm_setup_plane(display, pipe, j, parms, fencing);
 
 				if (type == TRANSITION_MODESET)
 					igt_output_override_mode(output, &override_mode);
 
 				atomic_commit(display, pipe, flags, (void *)(unsigned long) j, fencing);
-				drmHandleEvent(display->drm_fd, &drm_events);
+				wait_for_transition(display, pipe, nonblocking, fencing);
 
-				wm_setup_plane(display, pipe, i, parms);
+				wm_setup_plane(display, pipe, i, parms, fencing);
 				if (type == TRANSITION_MODESET)
 					igt_output_override_mode(output, NULL);
 
 				atomic_commit(display, pipe, flags, (void *)(unsigned long) i, fencing);
-				drmHandleEvent(display->drm_fd, &drm_events);
+				wait_for_transition(display, pipe, nonblocking, fencing);
 			}
 		}
 	}
@@ -600,7 +644,7 @@ static unsigned set_combinations(igt_display_t *display, unsigned mask, struct i
 		event_mask |= 1 << pipe;
 
 		for_each_valid_output_on_pipe(display, pipe, output) {
-			if (output->pending_crtc_idx_mask)
+			if (output->pending_pipe != PIPE_NONE)
 				continue;
 
 			mode = igt_output_get_mode(output);
@@ -619,22 +663,26 @@ static unsigned set_combinations(igt_display_t *display, unsigned mask, struct i
 	return event_mask;
 }
 
-static void refresh_primaries(igt_display_t *display)
+static void refresh_primaries(igt_display_t *display, int mask)
 {
 	enum pipe pipe;
 	igt_plane_t *plane;
 
-	for_each_pipe(display, pipe)
+	for_each_pipe(display, pipe) {
+		if (!((1 << pipe) & mask))
+			continue;
+
 		for_each_plane_on_pipe(display, pipe, plane)
-			if (plane->type == DRM_PLANE_TYPE_PRIMARY && plane->fb)
-				plane->fb_changed = true;
+			if (plane->type == DRM_PLANE_TYPE_PRIMARY)
+				igt_plane_set_position(plane, 0, 0);
+	}
 }
 
 static void collect_crcs_mask(igt_pipe_crc_t **pipe_crcs, unsigned mask, igt_crc_t *crcs)
 {
 	int i;
 
-	for (i = 0; i < I915_MAX_PIPES; i++) {
+	for (i = 0; i < IGT_MAX_PIPES; i++) {
 		if (!((1 << i) & mask))
 			continue;
 
@@ -650,7 +698,7 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 	struct igt_fb fbs[2];
 	int i, j;
 	unsigned iter_max = 1 << display->n_pipes;
-	igt_pipe_crc_t *pipe_crcs[I915_MAX_PIPES] = { 0 };
+	igt_pipe_crc_t *pipe_crcs[IGT_MAX_PIPES] = { 0 };
 	igt_output_t *output;
 	unsigned width = 0, height = 0;
 	bool skip_test = false;
@@ -678,7 +726,7 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 			pipe_crcs[i] = igt_pipe_crc_new(display->drm_fd, i, INTEL_PIPE_CRC_SOURCE_AUTO);
 
 		for_each_valid_output_on_pipe(display, i, output) {
-			if (output->pending_crtc_idx_mask)
+			if (output->pending_pipe != PIPE_NONE)
 				continue;
 
 			igt_output_set_pipe(output, i);
@@ -690,11 +738,11 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 			igt_plane_set_fb(plane, &fbs[1]);
 			igt_fb_set_size(&fbs[1], plane, mode->hdisplay, mode->vdisplay);
 			igt_plane_set_size(plane, mode->hdisplay, mode->vdisplay);
+
+			if (fencing)
+				igt_pipe_request_out_fence(&display->pipes[i]);
 		} else
 			igt_plane_set_fb(plane, NULL);
-
-		if(fencing)
-			igt_pipe_request_out_fence(&display->pipes[i]);
 	}
 
 	/*
@@ -707,7 +755,7 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 	igt_display_commit2(display, COMMIT_ATOMIC);
 
 	for (i = 0; i < iter_max; i++) {
-		igt_crc_t crcs[5][I915_MAX_PIPES];
+		igt_crc_t crcs[5][IGT_MAX_PIPES];
 		unsigned event_mask;
 
 		if (hweight32(i) > howmany)
@@ -736,7 +784,7 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 
 			collect_crcs_mask(pipe_crcs, j, crcs[1]);
 
-			refresh_primaries(display);
+			refresh_primaries(display, j);
 			commit_display(display, j, nonblocking);
 			collect_crcs_mask(pipe_crcs, j, crcs[2]);
 
@@ -747,14 +795,14 @@ static void run_modeset_tests(igt_display_t *display, int howmany, bool nonblock
 			commit_display(display, event_mask, nonblocking);
 			collect_crcs_mask(pipe_crcs, i, crcs[3]);
 
-			refresh_primaries(display);
+			refresh_primaries(display, i);
 			commit_display(display, i, nonblocking);
 			collect_crcs_mask(pipe_crcs, i, crcs[4]);
 
 			if (!is_i915_device(display->drm_fd))
 				continue;
 
-			for (int k = 0; k < I915_MAX_PIPES; k++) {
+			for (int k = 0; k < IGT_MAX_PIPES; k++) {
 				if (i & (1 << k)) {
 					igt_assert_crc_equal(&crcs[0][k], &crcs[3][k]);
 					igt_assert_crc_equal(&crcs[0][k], &crcs[4][k]);
@@ -784,7 +832,7 @@ cleanup:
 
 static void run_modeset_transition(igt_display_t *display, int requested_outputs, bool nonblocking, bool fencing)
 {
-	igt_output_t *outputs[I915_MAX_PIPES] = {};
+	igt_output_t *outputs[IGT_MAX_PIPES] = {};
 	int num_outputs = 0;
 	enum pipe pipe;
 
@@ -823,8 +871,6 @@ igt_main
 	igt_skip_on_simulation();
 
 	igt_fixture {
-		int valid_outputs = 0;
-
 		display.drm_fd = drm_open_driver_master(DRIVER_ANY);
 
 		kmstest_set_vt_graphics_mode();
@@ -833,10 +879,7 @@ igt_main
 
 		igt_require(display.is_atomic);
 
-		for_each_pipe_with_valid_output(&display, pipe, output)
-			valid_outputs++;
-
-		igt_require_f(valid_outputs, "no valid crtc/connector combinations found\n");
+		igt_display_require_output(&display);
 	}
 
 	igt_subtest("plane-primary-toggle-with-vblank-wait")
@@ -859,6 +902,14 @@ igt_main
 		for_each_pipe_with_valid_output(&display, pipe, output)
 			run_transition_test(&display, pipe, output, TRANSITION_PLANES, true, true);
 
+	igt_subtest("plane-use-after-nonblocking-unbind")
+		for_each_pipe_with_valid_output(&display, pipe, output)
+			run_transition_test(&display, pipe, output, TRANSITION_AFTER_FREE, true, false);
+
+	igt_subtest("plane-use-after-nonblocking-unbind-fencing")
+		for_each_pipe_with_valid_output(&display, pipe, output)
+			run_transition_test(&display, pipe, output, TRANSITION_AFTER_FREE, true, true);
+
 	igt_subtest("plane-all-modeset-transition")
 		for_each_pipe_with_valid_output(&display, pipe, output)
 			run_transition_test(&display, pipe, output, TRANSITION_MODESET, false, false);
@@ -871,7 +922,7 @@ igt_main
 		for_each_pipe_with_valid_output(&display, pipe, output)
 			run_transition_test(&display, pipe, output, TRANSITION_MODESET_DISABLE, false, false);
 
-	for (i = 1; i <= I915_MAX_PIPES; i++) {
+	for (i = 1; i <= IGT_MAX_PIPES; i++) {
 		igt_subtest_f("%ix-modeset-transitions", i)
 			run_modeset_transition(&display, i, false, false);
 
