@@ -1,5 +1,5 @@
 /*
- * Copyright © 2013 Intel Corporation
+ * Copyright © 2013, 2015 Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -25,11 +25,13 @@
  *
  */
 
-#include "igt.h"
+#include "config.h"
+
 #include <stdio.h>
 #include <stdint.h>
 #include <stdbool.h>
 #include <string.h>
+#include <ftw.h>
 
 #include <unistd.h>
 #include <fcntl.h>
@@ -43,6 +45,10 @@
 
 #include <drm.h>
 
+#include "igt.h"
+#include "igt_kmod.h"
+#include "igt_sysfs.h"
+#include "igt_debugfs.h"
 
 #define MSR_PC8_RES	0x630
 #define MSR_PC9_RES	0x631
@@ -52,8 +58,6 @@
 #define MAX_ENCODERS	32
 #define MAX_CRTCS	16
 
-#define POWER_DIR "/sys/devices/pci0000:00/0000:00:02.0/power"
-
 enum pc8_status {
 	PC8_ENABLED,
 	PC8_DISABLED
@@ -79,6 +83,7 @@ enum plane_type {
 #define USE_DPMS	8
 
 int drm_fd, msr_fd, pc8_status_fd;
+int debugfs;
 bool has_runtime_pm, has_pc8;
 struct mode_set_data ms_data;
 
@@ -111,6 +116,13 @@ struct modeset_params lpsp_mode_params;
 struct modeset_params non_lpsp_mode_params;
 struct modeset_params *default_mode_params;
 
+static int8_t *pm_data = NULL;
+
+static int modprobe(const char *driver)
+{
+	return igt_kmod_load(driver, "");
+}
+
 /* If the read fails, then the machine doesn't support PC8+ residencies. */
 static bool supports_pc8_plus_residencies(void)
 {
@@ -231,9 +243,8 @@ static bool init_modeset_params_for_type(struct mode_set_data *data,
 					 enum screen_type type)
 {
 	int i;
-	uint32_t connector_id = 0;
+	drmModeConnectorPtr connector = NULL;
 	drmModeModeInfoPtr mode = NULL;
-	cairo_t *cr;
 
 	for (i = 0; i < data->res->count_connectors; i++) {
 		drmModeConnectorPtr c = data->connectors[i];
@@ -247,24 +258,22 @@ static bool init_modeset_params_for_type(struct mode_set_data *data,
 			continue;
 
 		if (c->connection == DRM_MODE_CONNECTED && c->count_modes) {
-			connector_id = c->connector_id;
+			connector = c;
 			mode = &c->modes[0];
 			break;
 		}
 	}
 
-	if (!connector_id)
+	if (!connector)
 		return false;
 
-	igt_create_fb(drm_fd, mode->hdisplay, mode->vdisplay,
-		      DRM_FORMAT_XRGB8888, LOCAL_DRM_FORMAT_MOD_NONE,
-		      &params->fb);
-	cr = igt_get_cairo_ctx(drm_fd, &params->fb);
-	igt_paint_test_pattern(cr, mode->hdisplay, mode->vdisplay);
-	cairo_destroy(cr);
+	igt_create_pattern_fb(drm_fd, mode->hdisplay, mode->vdisplay,
+			      DRM_FORMAT_XRGB8888, LOCAL_DRM_FORMAT_MOD_NONE,
+			      &params->fb);
 
-	params->crtc_id = data->res->crtcs[0];
-	params->connector_id = connector_id;
+	params->crtc_id = kmstest_find_crtc_for_connector(drm_fd, data->res,
+							  connector, 0);
+	params->connector_id = connector->connector_id;
 	params->mode = mode;
 
 	return true;
@@ -370,7 +379,7 @@ static void init_mode_set_data(struct mode_set_data *data)
 	igt_assert(data->res->count_connectors <= MAX_CONNECTORS);
 
 	for (i = 0; i < data->res->count_connectors; i++) {
-		data->connectors[i] = drmModeGetConnector(drm_fd,
+		data->connectors[i] = drmModeGetConnectorCurrent(drm_fd,
 						data->res->connectors[i]);
 		data->edids[i] = get_connector_edid(data->connectors[i], i);
 	}
@@ -405,6 +414,8 @@ static void get_drm_info(struct compare_data *data)
 	igt_assert(data->res->count_crtcs <= MAX_CRTCS);
 
 	for (i = 0; i < data->res->count_connectors; i++) {
+		/* Don't use GetConnectorCurrent, we want to force a reprobe
+		 * here. */
 		data->connectors[i] = drmModeGetConnector(drm_fd,
 						data->res->connectors[i]);
 		data->edids[i] = get_connector_edid(data->connectors[i], i);
@@ -587,14 +598,14 @@ static int count_i2c_valid_edids(void)
 	DIR *dir;
 
 	struct dirent *dirent;
-	char full_name[32];
+	char full_name[PATH_MAX];
 
 	dir = opendir("/dev/");
 	igt_assert(dir);
 
 	while ((dirent = readdir(dir))) {
 		if (strncmp(dirent->d_name, "i2c-", 4) == 0) {
-			snprintf(full_name, 32, "/dev/%s", dirent->d_name);
+			sprintf(full_name, "/dev/%s", dirent->d_name);
 			fd = open(full_name, O_RDWR);
 			igt_assert_neq(fd, -1);
 			if (i2c_edid_is_valid(fd))
@@ -648,7 +659,7 @@ static void setup_pc8(void)
 		return;
 
 	/* Make sure our Kernel supports MSR and the module is loaded. */
-	igt_assert(system("modprobe -q msr > /dev/null 2>&1") != -1);
+	igt_require(modprobe("msr") == 0);
 
 	msr_fd = open("/dev/cpu/0/msr", O_RDONLY);
 	igt_assert_f(msr_fd >= 0,
@@ -658,51 +669,25 @@ static void setup_pc8(void)
 	if (!supports_pc8_plus_residencies())
 		return;
 
-	pc8_status_fd = igt_debugfs_open("i915_pc8_status", O_RDONLY);
+	pc8_status_fd = openat(debugfs, "i915_pc8_status", O_RDONLY);
 	if (pc8_status_fd == -1)
-		pc8_status_fd = igt_debugfs_open("i915_runtime_pm_status",
-						 O_RDONLY);
+		pc8_status_fd = openat(debugfs,
+				       "i915_runtime_pm_status", O_RDONLY);
 	igt_assert_f(pc8_status_fd >= 0,
 		     "Can't open /sys/kernel/debug/dri/0/i915_runtime_pm_status");
 
 	has_pc8 = true;
 }
 
-/* If we want to actually reach PC8+ states, we need to properly configure all
- * the devices on the system to allow this. This function will try to setup the
- * things we know we need, but won't scream in case anything fails: we don't
- * know which devices are present on your machine, so we can't really expect
- * anything, just try to help with the more common problems. */
-static void setup_non_graphics_runtime_pm(void)
-{
-	int fd, i;
-	char *file_name;
-
-	/* Disk runtime PM policies. */
-	file_name = malloc(PATH_MAX);
-	for (i = 0; ; i++) {
-
-		snprintf(file_name, PATH_MAX,
-			 "/sys/class/scsi_host/host%d/link_power_management_policy",
-			 i);
-
-		fd = open(file_name, O_WRONLY);
-		if (fd < 0)
-			break;
-
-		igt_assert(write(fd, "min_power\n", 10) == 10);
-		close(fd);
-	}
-	free(file_name);
-}
-
 static void setup_environment(void)
 {
 	drm_fd = drm_open_driver_master(DRIVER_INTEL);
+	debugfs = igt_debugfs_dir(drm_fd);
+	igt_require(debugfs != -1);
 
 	init_mode_set_data(&ms_data);
 
-	setup_non_graphics_runtime_pm();
+	pm_data = igt_pm_enable_sata_link_power_management();
 
 	has_runtime_pm = igt_setup_runtime_pm();
 	setup_pc8();
@@ -711,12 +696,19 @@ static void setup_environment(void)
 	igt_info("PC8 residency support: %d\n", has_pc8);
 
 	igt_require(has_runtime_pm);
+}
 
+static void restore_environment(void)
+{
+	igt_pm_restore_sata_link_power_management(pm_data);
+	free(pm_data);
 }
 
 static void teardown_environment(void)
 {
+	restore_environment();
 	fini_mode_set_data(&ms_data);
+	close(debugfs);
 	drmClose(drm_fd);
 	close(msr_fd);
 	if (has_pc8)
@@ -816,7 +808,7 @@ static void i2c_subtest_check_environment(void)
 	struct dirent *dirent;
 
 	/* Make sure the /dev/i2c-* files exist. */
-	igt_assert(system("modprobe -q i2c-dev > /dev/null 2>&1") != -1);
+	igt_require(modprobe("i2c-dev") == 0);
 
 	dev_dir = opendir("/dev");
 	igt_assert(dev_dir);
@@ -842,63 +834,40 @@ static void i2c_subtest(void)
 	enable_one_screen(&ms_data);
 }
 
-static void read_full_file(const char *name)
+static int read_entry(const char *filepath,
+		      const struct stat *info,
+		      const int typeflag,
+		      struct FTW *pathinfo)
 {
-	int rc, fd;
-	char buf[128];
-
-	igt_assert_f(wait_for_suspended(), "File: %s\n", name);
-
-	fd = open(name, O_RDONLY);
-	if (fd < 0)
-		return;
-
-	do {
-		rc = read(fd, buf, ARRAY_SIZE(buf));
-	} while (rc == ARRAY_SIZE(buf));
-
-	rc = close(fd);
-	igt_assert_eq(rc, 0);
-
-	igt_assert_f(wait_for_suspended(), "File: %s\n", name);
-}
-
-static void read_files_from_dir(const char *name, int level)
-{
-	DIR *dir;
-	struct dirent *dirent;
-	char *full_name;
+	char buf[4096];
+	int fd;
 	int rc;
 
-	dir = opendir(name);
-	igt_assert(dir);
+	igt_assert_f(wait_for_suspended(), "Before opening: %s (%s)\n",
+		     filepath + pathinfo->base, filepath);
 
-	full_name = malloc(PATH_MAX);
-
-	igt_assert_lt(level, 128);
-
-	while ((dirent = readdir(dir))) {
-		struct stat stat_buf;
-
-		if (strcmp(dirent->d_name, ".") == 0)
-			continue;
-		if (strcmp(dirent->d_name, "..") == 0)
-			continue;
-
-		snprintf(full_name, PATH_MAX, "%s/%s", name, dirent->d_name);
-
-		rc = lstat(full_name, &stat_buf);
-		igt_assert_eq(rc, 0);
-
-		if (S_ISDIR(stat_buf.st_mode))
-			read_files_from_dir(full_name, level + 1);
-
-		if (S_ISREG(stat_buf.st_mode))
-			read_full_file(full_name);
+	fd = open(filepath, O_RDONLY | O_NONBLOCK);
+	if (fd < 0) {
+		igt_debug("Failed to open '%s': %m\n", filepath);
+		return 0;
 	}
 
-	free(full_name);
-	closedir(dir);
+	do {
+		rc = read(fd, buf, sizeof(buf));
+	} while (rc == sizeof(buf));
+
+	close(fd);
+
+	igt_assert_f(wait_for_suspended(), "After closing: %s (%s)\n",
+		     filepath + pathinfo->base, filepath);
+
+	return 0;
+}
+
+static void walk_fs(char *path)
+{
+	disable_all_screens_and_wait(&ms_data);
+	nftw(path, read_entry, 20, FTW_PHYS | FTW_MOUNT);
 }
 
 /* This test will probably pass, with a small chance of hanging the machine in
@@ -906,31 +875,21 @@ static void read_files_from_dir(const char *name, int level)
  * errors, so a "pass" here should be confirmed by a check on dmesg. */
 static void debugfs_read_subtest(void)
 {
-	const char *path = "/sys/kernel/debug/dri/0";
-	DIR *dir;
+	char path[256];
 
-	dir = opendir(path);
-	igt_require_f(dir, "Can't open the debugfs directory\n");
-	closedir(dir);
-
-	disable_all_screens_and_wait(&ms_data);
-
-	read_files_from_dir(path, 0);
+	igt_require_f(igt_debugfs_path(drm_fd, path, sizeof(path)),
+		      "Can't find the debugfs directory\n");
+	walk_fs(path);
 }
 
 /* Read the comment on debugfs_read_subtest(). */
 static void sysfs_read_subtest(void)
 {
-	const char *path = "/sys/devices/pci0000:00/0000:00:02.0";
-	DIR *dir;
+	char path[80];
 
-	dir = opendir(path);
-	igt_require_f(dir, "Can't open the sysfs directory\n");
-	closedir(dir);
-
-	disable_all_screens_and_wait(&ms_data);
-
-	read_files_from_dir(path, 0);
+	igt_require_f(igt_sysfs_path(drm_fd, path, sizeof(path), NULL),
+		      "Can't find the sysfs directory\n");
+	walk_fs(path);
 }
 
 /* Make sure we don't suspend when we have the i915_forcewake_user file open. */
@@ -942,7 +901,7 @@ static void debugfs_forcewake_user_subtest(void)
 
 	disable_all_screens_and_wait(&ms_data);
 
-	fd = igt_open_forcewake_handle();
+	fd = igt_open_forcewake_handle(drm_fd);
 	igt_require(fd >= 0);
 
 	if (has_runtime_pm) {
@@ -1360,15 +1319,13 @@ static void reg_read_ioctl_subtest(void)
 
 static bool device_in_pci_d3(void)
 {
-	struct pci_device *pci_dev;
-	int rc;
 	uint16_t val;
+	int rc;
 
-	pci_dev = intel_get_pci_device();
-
-	rc = pci_device_cfg_read_u16(pci_dev, &val, 0xd4);
+	rc = pci_device_cfg_read_u16(intel_get_pci_device(), &val, 0xd4);
 	igt_assert_eq(rc, 0);
 
+	igt_debug("%s: PCI D3 state=%d\n", __func__, val & 0x3);
 	return (val & 0x3) == 0x3;
 }
 
@@ -1377,11 +1334,9 @@ static void pci_d3_state_subtest(void)
 	igt_require(has_runtime_pm);
 
 	disable_all_screens_and_wait(&ms_data);
-
-	igt_assert(device_in_pci_d3());
+	igt_assert(igt_wait(device_in_pci_d3(), 2000, 100));
 
 	enable_one_screen_and_wait(&ms_data);
-
 	igt_assert(!device_in_pci_d3());
 }
 
@@ -1396,7 +1351,7 @@ static void __attribute__((noreturn)) stay_subtest(void)
 static void system_suspend_subtest(void)
 {
 	disable_all_screens_and_wait(&ms_data);
-	igt_system_suspend_autoresume();
+	igt_system_suspend_autoresume(SUSPEND_STATE_MEM, SUSPEND_TEST_NONE);
 	igt_assert(wait_for_suspended());
 }
 
@@ -1428,7 +1383,7 @@ static void system_suspend_execbuf_subtest(void)
 	i915_execbuffer2_set_context_id(execbuf, 0);
 
 	disable_all_screens_and_wait(&ms_data);
-	igt_system_suspend_autoresume();
+	igt_system_suspend_autoresume(SUSPEND_STATE_MEM, SUSPEND_TEST_NONE);
 	igt_assert(wait_for_suspended());
 
 	for (i = 0; i < 20; i++) {
@@ -1442,7 +1397,7 @@ static void system_suspend_execbuf_subtest(void)
 static void system_suspend_modeset_subtest(void)
 {
 	disable_all_screens_and_wait(&ms_data);
-	igt_system_suspend_autoresume();
+	igt_system_suspend_autoresume(SUSPEND_STATE_MEM, SUSPEND_TEST_NONE);
 	igt_assert(wait_for_suspended());
 
 	enable_one_screen_and_wait(&ms_data);
@@ -1603,7 +1558,6 @@ static void test_one_plane(bool dpms, uint32_t plane_id,
 
 	disable_all_screens_and_wait(&ms_data);
 
-	igt_require(default_mode_params);
 	crtc_id = default_mode_params->crtc_id;
 
 	switch (plane_type) {
@@ -1688,9 +1642,13 @@ static void test_one_plane(bool dpms, uint32_t plane_id,
 /* This one also triggered WARNs on our driver at some point in time. */
 static void planes_subtest(bool universal, bool dpms)
 {
-	int i, rc, planes_tested = 0;
+	int i, rc, planes_tested = 0, crtc_idx;
 	drmModePlaneResPtr planes;
 
+	igt_require(default_mode_params);
+	crtc_idx = kmstest_get_crtc_idx(ms_data.res,
+					default_mode_params->crtc_id);
+
 	if (universal) {
 		rc = drmSetClientCap(drm_fd, DRM_CLIENT_CAP_UNIVERSAL_PLANES,
 				     1);
@@ -1704,9 +1662,7 @@ static void planes_subtest(bool universal, bool dpms)
 		plane = drmModeGetPlane(drm_fd, planes->planes[i]);
 		igt_assert(plane);
 
-		/* We just pick the first CRTC on the list, so we can test for
-		 * 0x1 as the index. */
-		if (plane->possible_crtcs & 0x1) {
+		if (plane->possible_crtcs & (1 << crtc_idx)) {
 			enum plane_type type;
 
 			type = universal ? get_plane_type(plane->plane_id) :
@@ -1792,7 +1748,8 @@ static void pm_test_caching(void)
 	uint32_t handle;
 	uint8_t *gem_buf;
 
-	uint32_t i, got_caching;
+	uint32_t i;
+	uint32_t default_cache_level;
 	uint32_t gtt_obj_max_size = (16 * 1024);
 	uint32_t cache_levels[3] = {
 		I915_CACHING_NONE,
@@ -1800,42 +1757,34 @@ static void pm_test_caching(void)
 		I915_CACHING_DISPLAY,           /* eDRAM caching */
 	};
 
+	disable_all_screens(&ms_data);
 
 	handle = gem_create(drm_fd, gtt_obj_max_size);
+	default_cache_level = gem_get_caching(drm_fd, handle);
 	gem_buf = gem_mmap__gtt(drm_fd, handle, gtt_obj_max_size, PROT_WRITE);
 
 	for (i = 0; i < ARRAY_SIZE(cache_levels); i++) {
+		igt_assert(wait_for_suspended());
+		gem_set_caching(drm_fd, handle, default_cache_level);
+
+		/* Ensure we bind the vma into the GGTT */
 		memset(gem_buf, 16 << i, gtt_obj_max_size);
 
-		disable_all_screens_and_wait(&ms_data);
-
-		igt_debug("Setting cache level %u\n", cache_levels[i]);
-
-		gem_set_caching(drm_fd, handle, cache_levels[i]);
-
-		got_caching = gem_get_caching(drm_fd, handle);
-
-		igt_debug("Got back %u\n", got_caching);
-
-		/*
-		 * Allow fall-back to CACHING_NONE in case the platform does
-		 * not support it.
+		/* Now try changing the cache-level on the bound object.
+		 * This will either unlikely unbind the object from the GGTT,
+		 * or more likely just change the PTEs inside the GGTT. Either
+		 * way the driver must take the rpm wakelock around the GSM
+		 * access.
 		 */
-		if (cache_levels[i] == I915_CACHING_DISPLAY)
-			igt_assert(got_caching == I915_CACHING_NONE ||
-				   got_caching == I915_CACHING_DISPLAY);
-		else
-			igt_assert(got_caching == cache_levels[i]);
-
-		enable_one_screen_and_wait(&ms_data);
+		igt_debug("Setting cache level %u\n", cache_levels[i]);
+		igt_assert(wait_for_suspended());
+		gem_set_caching(drm_fd, handle, cache_levels[i]);
 	}
 
 	igt_assert(munmap(gem_buf, gtt_obj_max_size) == 0);
 	gem_close(drm_fd, handle);
 }
 
-
-
 static void fences_subtest(bool dpms)
 {
 	int i;
@@ -1892,14 +1841,14 @@ static void fences_subtest(bool dpms)
 	igt_assert(munmap(buf_ptr, params.fb.size) == 0);
 }
 
-int rounds = 40;
+int rounds = 10;
 bool stay = false;
 
 static int opt_handler(int opt, int opt_index, void *data)
 {
 	switch (opt) {
-	case 'q':
-		rounds = 10;
+	case 'l':
+		rounds = 50;
 		break;
 	case 's':
 		stay = true;
@@ -1914,10 +1863,10 @@ static int opt_handler(int opt, int opt_index, void *data)
 int main(int argc, char *argv[])
 {
 	const char *help_str =
-	       "  --quick\t\tMake the stress-tests not stressful, for quick regression testing.\n"
+	       "  --stress\t\tMake the stress-tests more stressful.\n"
 	       "  --stay\t\tDisable all screen and try to go into runtime pm. Useful for debugging.";
 	static struct option long_options[] = {
-		{"quick", 0, 0, 'q'},
+		{"stress", 0, 0, 'l'},
 		{"stay", 0, 0, 's'},
 		{ 0, 0, 0, 0 }
 	};
