@@ -21,16 +21,16 @@
  * IN THE SOFTWARE.
  */
 
-#define _GNU_SOURCE /* for RTLD_NEXT */
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdint.h>
 #include <stdbool.h>
+#include <signal.h>
 #include <stdarg.h>
 #include <fcntl.h>
 #include <sys/types.h>
+#include <sys/sysmacros.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <unistd.h>
@@ -42,15 +42,21 @@
 #include "intel_aub.h"
 #include "intel_chipset.h"
 
-static int (*libc_close)(int fd);
-static int (*libc_ioctl)(int fd, unsigned long request, void *argp);
+#ifndef ARRAY_SIZE
+#define ARRAY_SIZE(x) (sizeof(x)/sizeof((x)[0]))
+#endif
+
+static int close_init_helper(int fd);
+static int ioctl_init_helper(int fd, unsigned long request, ...);
+
+static int (*libc_close)(int fd) = close_init_helper;
+static int (*libc_ioctl)(int fd, unsigned long request, ...) = ioctl_init_helper;
 
 static int drm_fd = -1;
-static char *filename;
-static FILE *file;
+static char *filename = NULL;
+static FILE *files[2] = { NULL, NULL };
 static int gen = 0;
 static int verbose = 0;
-static const uint32_t gtt_size = 0x10000;
 static bool device_override;
 static uint32_t device;
 
@@ -94,6 +100,10 @@ struct drm_i915_gem_userptr {
 #define IS_USERPTR(p) ((uintptr_t) (p) & USERPTR_FLAG)
 #define GET_PTR(p) ( (void *) ((uintptr_t) p & ~(uintptr_t) 1) )
 
+#ifndef I915_EXEC_BATCH_FIRST
+#define I915_EXEC_BATCH_FIRST (1 << 18)
+#endif
+
 static void __attribute__ ((format(__printf__, 2, 3)))
 fail_if(int cond, const char *format, ...)
 {
@@ -106,7 +116,18 @@ fail_if(int cond, const char *format, ...)
 	vfprintf(stderr, format, args);
 	va_end(args);
 
-	exit(1);
+	raise(SIGTRAP);
+}
+
+static struct bo *
+get_bo(uint32_t handle)
+{
+	struct bo *bo;
+
+	fail_if(handle >= MAX_BO_COUNT, "bo handle too large\n");
+	bo = &bos[handle];
+
+	return bo;
 }
 
 static inline uint32_t
@@ -124,42 +145,85 @@ align_u64(uint64_t v, uint64_t a)
 static void
 dword_out(uint32_t data)
 {
-	fwrite(&data, 1, 4, file);
+	for (int i = 0; i < ARRAY_SIZE (files); i++) {
+		if (files[i] == NULL)
+			continue;
+
+		fail_if(fwrite(&data, 1, 4, files[i]) == 0,
+			"Writing to output failed\n");
+	}
 }
 
 static void
 data_out(const void *data, size_t size)
 {
-	fwrite(data, 1, size, file);
+	if (size == 0)
+		return;
+
+	for (int i = 0; i < ARRAY_SIZE (files); i++) {
+		if (files[i] == NULL)
+			continue;
+
+		fail_if(fwrite(data, 1, size, files[i]) == 0,
+			"Writing to output failed\n");
+	}
+}
+
+static uint32_t
+gtt_entry_size(void)
+{
+	return gen >= 8 ? 8 : 4;
+}
+
+static uint32_t
+gtt_size(void)
+{
+	/* Enough for 64MB assuming 4kB pages. */
+	const unsigned entries = 0x4000;
+	return entries * gtt_entry_size();
 }
 
 static void
 write_header(void)
 {
+	char app_name[8 * 4];
+	char comment[16];
+	int comment_len, comment_dwords, dwords;
 	uint32_t entry = 0x200003;
 
+	comment_len = snprintf(comment, sizeof(comment), "PCI-ID=0x%x", device);
+	comment_dwords = ((comment_len + 3) / 4);
+
 	/* Start with a (required) version packet. */
-	dword_out(CMD_AUB_HEADER | (13 - 2));
+	dwords = 13 + comment_dwords;
+	dword_out(CMD_AUB_HEADER | (dwords - 2));
 	dword_out((4 << AUB_HEADER_MAJOR_SHIFT) |
 		  (0 << AUB_HEADER_MINOR_SHIFT));
-	for (int i = 0; i < 8; i++)
-		dword_out(0); /* app name */
+
+	/* Next comes a 32-byte application name. */
+	strncpy(app_name, program_invocation_short_name, sizeof(app_name));
+	app_name[sizeof(app_name) - 1] = 0;
+	data_out(app_name, sizeof(app_name));
 
 	dword_out(0); /* timestamp */
 	dword_out(0); /* timestamp */
-	dword_out(0); /* comment len */
+	dword_out(comment_len);
+	data_out(comment, comment_dwords * 4);
 
-	/* Set up the GTT. The max we can handle is 256M */
+	/* Set up the GTT. The max we can handle is 64M */
 	dword_out(CMD_AUB_TRACE_HEADER_BLOCK | ((gen >= 8 ? 6 : 5) - 2));
 	dword_out(AUB_TRACE_MEMTYPE_GTT_ENTRY |
 		  AUB_TRACE_TYPE_NOTYPE | AUB_TRACE_OP_DATA_WRITE);
 	dword_out(0); /* subtype */
 	dword_out(0); /* offset */
-	dword_out(gtt_size); /* size */
+	dword_out(gtt_size()); /* size */
 	if (gen >= 8)
 		dword_out(0);
-	for (uint32_t i = 0; i < gtt_size; i += 4, entry += 0x1000)
-		dword_out(entry);
+	for (uint32_t i = 0; i * gtt_entry_size() < gtt_size(); i++) {
+		dword_out(entry + 0x1000 * i);
+		if (gen >= 8)
+			dword_out(0);
+	}
 }
 
 /**
@@ -200,10 +264,37 @@ aub_write_trace_block(uint32_t type, void *virtual, uint32_t size, uint64_t gtt_
 	}
 }
 
+static void
+write_reloc(void *p, uint64_t v)
+{
+	if (gen >= 8) {
+		/* From the Broadwell PRM Vol. 2a,
+		 * MI_LOAD_REGISTER_MEM::MemoryAddress:
+		 *
+		 *	"This field specifies the address of the memory
+		 *	location where the register value specified in the
+		 *	DWord above will read from.  The address specifies
+		 *	the DWord location of the data. Range =
+		 *	GraphicsVirtualAddress[63:2] for a DWord register
+		 *	GraphicsAddress [63:48] are ignored by the HW and
+		 *	assumed to be in correct canonical form [63:48] ==
+		 *	[47]."
+		 *
+		 * In practice, this will always mean the top bits are zero
+		 * because of the GTT size limitation of the aubdump tool.
+		 */
+		const int shift = 63 - 47;
+		*(uint64_t *)p = (((int64_t)v) << shift) >> shift;
+	} else {
+		*(uint32_t *)p = v;
+	}
+}
+
 static void
 aub_dump_ringbuffer(uint64_t batch_offset, uint64_t offset, int ring_flag)
 {
 	uint32_t ringbuffer[4096];
+	unsigned aub_mi_bbs_len;
 	int ring = AUB_TRACE_TYPE_RING_PRB0; /* The default ring */
 	int ring_count = 0;
 
@@ -214,14 +305,11 @@ aub_dump_ringbuffer(uint64_t batch_offset, uint64_t offset, int ring_flag)
 
 	/* Make a ring buffer to execute our batchbuffer. */
 	memset(ringbuffer, 0, sizeof(ringbuffer));
-	if (gen >= 8) {
-		ringbuffer[ring_count++] = AUB_MI_BATCH_BUFFER_START | (3 - 2);
-		ringbuffer[ring_count++] = batch_offset;
-		ringbuffer[ring_count++] = batch_offset >> 32;
-	} else {
-		ringbuffer[ring_count++] = AUB_MI_BATCH_BUFFER_START;
-		ringbuffer[ring_count++] = batch_offset;
-	}
+
+	aub_mi_bbs_len = gen >= 8 ? 3 : 2;
+	ringbuffer[ring_count] = AUB_MI_BATCH_BUFFER_START | (aub_mi_bbs_len - 2);
+	write_reloc(&ringbuffer[ring_count + 1], batch_offset);
+	ring_count += aub_mi_bbs_len;
 
 	/* Write out the ring.  This appears to trigger execution of
 	 * the ring in the simulator.
@@ -247,7 +335,6 @@ relocate_bo(struct bo *bo, const struct drm_i915_gem_execbuffer2 *execbuffer2,
 	const struct drm_i915_gem_relocation_entry *relocs =
 		(const struct drm_i915_gem_relocation_entry *) (uintptr_t) obj->relocs_ptr;
 	void *relocated;
-	uint32_t *dw;
 	int handle;
 
 	relocated = malloc(bo->size);
@@ -261,8 +348,8 @@ relocate_bo(struct bo *bo, const struct drm_i915_gem_execbuffer2 *execbuffer2,
 		else
 			handle = relocs[i].target_handle;
 
-		dw = (uint32_t*)(((char *) relocated) + relocs[i].offset);
-		*dw = bos[handle].offset + relocs[i].delta;
+		write_reloc(((char *)relocated) + relocs[i].offset,
+			    get_bo(handle)->offset + relocs[i].delta);
 	}
 
 	return relocated;
@@ -316,7 +403,7 @@ dump_execbuffer2(int fd, struct drm_i915_gem_execbuffer2 *execbuffer2)
 	struct drm_i915_gem_exec_object2 *exec_objects =
 		(struct drm_i915_gem_exec_object2 *) (uintptr_t) execbuffer2->buffers_ptr;
 	uint32_t ring_flag = execbuffer2->flags & I915_EXEC_RING_MASK;
-	uint32_t offset = gtt_size;
+	uint32_t offset = gtt_size();
 	struct drm_i915_gem_exec_object2 *obj;
 	struct bo *bo, *batch_bo;
 	void *data;
@@ -336,22 +423,43 @@ dump_execbuffer2(int fd, struct drm_i915_gem_execbuffer2 *execbuffer2)
 			       filename, device, gen);
 	}
 
+	if (verbose)
+		printf("Dumping execbuffer2:\n");
+
 	for (uint32_t i = 0; i < execbuffer2->buffer_count; i++) {
 		obj = &exec_objects[i];
-		bo = &bos[obj->handle];
+		bo = get_bo(obj->handle);
 
-		bo->offset = offset;
-		offset = align_u32(offset + bo->size + 4095, 4096);
+		/* If bo->size == 0, this means they passed us an invalid
+		 * buffer.  The kernel will reject it and so should we.
+		 */
+		if (bo->size == 0) {
+			if (verbose)
+				printf("BO #%d is invalid!\n", obj->handle);
+			return;
+		}
 
-		if (bo->map == NULL)
+		if (verbose)
+			printf("BO #%d (%dB) @ 0x%x\n", obj->handle, bo->size, offset);
+
+		if (obj->flags & EXEC_OBJECT_PINNED) {
+			bo->offset = obj->offset;
+		} else {
+			bo->offset = offset;
+			offset = align_u32(offset + bo->size + 4095, 4096);
+		}
+
+		if (bo->map == NULL && bo->size > 0)
 			bo->map = gem_mmap(fd, obj->handle, 0, bo->size);
 		fail_if(bo->map == MAP_FAILED, "intel_aubdump: bo mmap failed\n");
 	}
 
-	batch_bo = &bos[exec_objects[execbuffer2->buffer_count - 1].handle];
+	int batch_index = (execbuffer2->flags & I915_EXEC_BATCH_FIRST) ? 0 :
+			  execbuffer2->buffer_count - 1;
+	batch_bo = get_bo(exec_objects[batch_index].handle);
 	for (uint32_t i = 0; i < execbuffer2->buffer_count; i++) {
 		obj = &exec_objects[i];
-		bo = &bos[obj->handle];
+		bo = get_bo(obj->handle);
 
 		if (obj->relocation_count > 0)
 			data = relocate_bo(bo, execbuffer2, obj);
@@ -373,7 +481,10 @@ dump_execbuffer2(int fd, struct drm_i915_gem_execbuffer2 *execbuffer2)
 	aub_dump_ringbuffer(batch_bo->offset + execbuffer2->batch_start_offset,
 			    offset, ring_flag);
 
-	fflush(file);
+	for (int i = 0; i < ARRAY_SIZE(files); i++) {
+		if (files[i] != NULL)
+			fflush(files[i]);
+	}
 }
 
 static void
@@ -382,6 +493,7 @@ add_new_bo(int handle, uint64_t size, void *map)
 	struct bo *bo = &bos[handle];
 
 	fail_if(handle >= MAX_BO_COUNT, "intel_aubdump: bo handle out of range\n");
+	fail_if(size == 0, "intel_aubdump: bo size is invalid\n");
 
 	bo->size = size;
 	bo->map = map;
@@ -390,10 +502,11 @@ add_new_bo(int handle, uint64_t size, void *map)
 static void
 remove_bo(int handle)
 {
-	struct bo *bo = &bos[handle];
+	struct bo *bo = get_bo(handle);
 
 	if (bo->map && !IS_USERPTR(bo->map))
 		munmap(bo->map, bo->size);
+	bo->size = 0;
 	bo->map = NULL;
 }
 
@@ -406,6 +519,88 @@ close(int fd)
 	return libc_close(fd);
 }
 
+static FILE *
+launch_command(char *command)
+{
+	int i = 0, fds[2];
+	char **args = calloc(strlen(command), sizeof(char *));
+	char *iter = command;
+
+	args[i++] = iter = command;
+
+	while ((iter = strstr(iter, ",")) != NULL) {
+		*iter = '\0';
+		iter += 1;
+		args[i++] = iter;
+	}
+
+	if (pipe(fds) == -1)
+		return NULL;
+
+	switch (fork()) {
+	case 0:
+		dup2(fds[0], 0);
+		fail_if(execvp(args[0], args) == -1,
+			"intel_aubdump: failed to launch child command\n");
+		return NULL;
+
+	default:
+		free(args);
+		return fdopen(fds[1], "w");
+
+	case -1:
+		return NULL;
+	}
+}
+
+static void
+maybe_init(void)
+{
+	static bool initialized = false;
+	FILE *config;
+	char *key, *value;
+
+	if (initialized)
+		return;
+
+	initialized = true;
+
+	config = fdopen(3, "r");
+	while (fscanf(config, "%m[^=]=%m[^\n]\n", &key, &value) != EOF) {
+		if (!strcmp(key, "verbose")) {
+			verbose = 1;
+		} else if (!strcmp(key, "device")) {
+			fail_if(sscanf(value, "%i", &device) != 1,
+				"intel_aubdump: failed to parse device id '%s'",
+				value);
+			device_override = true;
+		} else if (!strcmp(key, "file")) {
+			filename = strdup(value);
+			files[0] = fopen(filename, "w+");
+			fail_if(files[0] == NULL,
+				"intel_aubdump: failed to open file '%s'\n",
+				filename);
+		} else if (!strcmp(key,  "command")) {
+			files[1] = launch_command(value);
+			fail_if(files[1] == NULL,
+				"intel_aubdump: failed to launch command '%s'\n",
+				value);
+		} else {
+			fprintf(stderr, "intel_aubdump: unknown option '%s'\n", key);
+		}
+
+		free(key);
+		free(value);
+	}
+	fclose(config);
+
+	bos = calloc(MAX_BO_COUNT, sizeof(bos[0]));
+	fail_if(bos == NULL, "intel_aubdump: out of memory\n");
+}
+
+#define LOCAL_IOCTL_I915_GEM_EXECBUFFER2_WR \
+    DRM_IOWR(DRM_COMMAND_BASE + DRM_I915_GEM_EXECBUFFER2, struct drm_i915_gem_execbuffer2)
+
 int
 ioctl(int fd, unsigned long request, ...)
 {
@@ -427,6 +622,8 @@ ioctl(int fd, unsigned long request, ...)
 	}
 
 	if (fd == drm_fd) {
+		maybe_init();
+
 		switch (request) {
 		case DRM_IOCTL_I915_GETPARAM: {
 			struct drm_i915_getparam *getparam = argp;
@@ -458,7 +655,8 @@ ioctl(int fd, unsigned long request, ...)
 			return libc_ioctl(fd, request, argp);
 		}
 
-		case DRM_IOCTL_I915_GEM_EXECBUFFER2: {
+		case DRM_IOCTL_I915_GEM_EXECBUFFER2:
+		case LOCAL_IOCTL_I915_GEM_EXECBUFFER2_WR: {
 			dump_execbuffer2(fd, argp);
 			if (device_override)
 				return 0;
@@ -527,35 +725,43 @@ ioctl(int fd, unsigned long request, ...)
 	}
 }
 
-static void __attribute__ ((constructor))
+static void
 init(void)
 {
-	const char *args = getenv("INTEL_AUBDUMP_ARGS");
-
 	libc_close = dlsym(RTLD_NEXT, "close");
 	libc_ioctl = dlsym(RTLD_NEXT, "ioctl");
 	fail_if(libc_close == NULL || libc_ioctl == NULL,
 		"intel_aubdump: failed to get libc ioctl or close\n");
+}
 
-	if (sscanf(args, "verbose=%d;file=%m[^;];device=%i",
-		   &verbose, &filename, &device) != 3)
-		filename = strdup("intel.aub");
-	fail_if(filename == NULL, "intel_aubdump: out of memory\n");
+static int
+close_init_helper(int fd)
+{
+	init();
+	return libc_close(fd);
+}
 
-	if (device)
-		device_override = true;
+static int
+ioctl_init_helper(int fd, unsigned long request, ...)
+{
+	va_list args;
+	void *argp;
 
-	bos = malloc(MAX_BO_COUNT * sizeof(bos[0]));
-	fail_if(bos == NULL, "intel_aubdump: out of memory\n");
+	va_start(args, request);
+	argp = va_arg(args, void *);
+	va_end(args);
 
-	file = fopen(filename, "w+");
-	fail_if(file == NULL, "intel_aubdump: failed to open file '%s'\n", filename);
+	init();
+	return libc_ioctl(fd, request, argp);
 }
 
 static void __attribute__ ((destructor))
 fini(void)
 {
 	free(filename);
-	fclose(file);
+	for (int i = 0; i < ARRAY_SIZE(files); i++) {
+		if (files[i] != NULL)
+			fclose(files[i]);
+	}
 	free(bos);
 }
